from qgis.PyQt.QtCore import (QCoreApplication, QT_TRANSLATE_NOOP)
from qgis.core import (
  QgsProcessing,
  QgsProcessingParameterFeatureSource,
  QgsProcessingParameterFeatureSink,
  QgsProcessingParameterCrs,
  QgsProcessingParameterDistance,
  QgsProcessingParameterFeatureSource,
  QgsFeatureRequest,
  QgsFeature,
  QgsWkbTypes,
  QgsVectorLayer,
  QgsProperty
  )

from qgis import processing

from .receiverabstract import receiverabstract
import numpy as np
import sys

class receivercrosssection(receiverabstract):
  PARAMETERS = { 
    "RECEIVER_LINE": {
      "crs_referrence": True, # this parameter is used as CRS referrence
      "ui_func": QgsProcessingParameterFeatureSource,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("receivercrosssection","Receiver line layer"),
        "types": [QgsProcessing.TypeVectorLine]
      }
    },    
    
    "TARGET_CRS": {
      "ui_func": QgsProcessingParameterCrs,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("receivercrosssection","Target CRS (Cartesian coordinates)")
      }
    },
    "INTERVAL": {
      "ui_func": QgsProcessingParameterDistance,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("receivercrosssection","Horizontal / vertical interval between receivers"),
        "defaultValue": 2.0,
        "parentParameterName": "TARGET_CRS"
      }
    },
    "HEIGHT_MIN": {
      "ui_func": QgsProcessingParameterDistance,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("receivercrosssection","Minimum height of the receivers"),
        "defaultValue": 2.0,
        "parentParameterName": "TARGET_CRS"
      }
    },
    "HEIGHT_MAX": {
      "ui_func": QgsProcessingParameterDistance,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("receivercrosssection","Maximum height of the receivers"),
        "defaultValue": 20.0,
        "parentParameterName": "TARGET_CRS"
      }
    },
    "OUTPUT": {
      "ui_func": QgsProcessingParameterFeatureSink,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("receivercrosssection","Receivers of crosssection" )
      }
    }

  }
  
    
  def initAlgorithm(self, config):
    (rect, target_crs) = self.getExtentAndCrsUsingCanvas()
    self.PARAMETERS["TARGET_CRS"]["ui_args"]["defaultValue"] = target_crs.authid()
    self.initParameters()

  def processAlgorithm(self, parameters, context, feedback):   
    target_crs = self.parameterAsCrs(parameters, "TARGET_CRS", context)
    
    # list from HEIGHT_MIN to HEIGHT_MAX with INTERVAL
    receiver_heights = np.arange(
      self.parameterAsDouble(parameters, "HEIGHT_MIN", context),
      self.parameterAsDouble(parameters, "HEIGHT_MAX", context),
      self.parameterAsDouble(parameters, "INTERVAL", context)
    )
    
    empty_receiver = processing.run(
      "hrisk:initreceiver",
      {
        "TARGET_CRS": target_crs,
        "OUTPUT": "TEMPORARY_OUTPUT"
      }
    )["OUTPUT"]
    
    # check whether the target CRS is x-y coordinates
    if target_crs.isGeographic() or target_crs.authid() == "EPSG:3857":
      sys.exit(self.tr("The Target CRS is NOT a Cartesian Coordinate System"))
      
    line_layer_raw = processing.run(
      "native:reprojectlayer", 
      {
        "INPUT": self.parameterAsSource(parameters, "RECEIVER_LINE", context).materialize(QgsFeatureRequest(), feedback),
        "TARGET_CRS": target_crs,
        "OUTPUT": "TEMPORARY_OUTPUT"
      }
    )["OUTPUT"]      
            
    line_layer_single = processing.run(
      "native:multiparttosingleparts",
      {
        "INPUT": line_layer_raw,
        "OUTPUT": "TEMPORARY_OUTPUT"
      }
    )["OUTPUT"]
    
    
    # set sink and add features with values
    (sink, dest_id) = self.parameterAsSink(
      parameters, "OUTPUT", context,
      empty_receiver.fields(), QgsWkbTypes.PointZ, target_crs
    )
      
    pk = 0
    for ft in line_layer_single.getFeatures():
      line_layer = QgsVectorLayer(f"LineString?crs={target_crs.authid()}", "temp", "memory")
      line_layer.dataProvider().addFeatures([ft])
      pnts_on_line = processing.run(
        "native:pointsalonglines",
        {
          "INPUT": line_layer,
          "DISTANCE": self.parameterAsDouble(parameters, "INTERVAL", context),
          "START_OFFSET": 0,
          "END_OFFSET": 0,
          "OUTPUT": "TEMPORARY_OUTPUT"
        }
      )["OUTPUT"]
      
      for height in receiver_heights:
        
        # set z value
        pnt_z = processing.run(
          "native:setzvalue",
          {
            "INPUT": pnts_on_line,
            "Z_VALUE": float(height),
            "OUTPUT": "memory:dem"
          }
        )["OUTPUT"]
        
        for pnt in pnt_z.getFeatures():
        
          ft_new = QgsFeature(empty_receiver.fields())
          ft_new.setGeometry(pnt.geometry())
          ft_new["PK"] = pk
          sink.addFeature(ft_new)
          pk += 1
      
    return {"OUTPUT": dest_id}
    
  # Post processing; append layers
  def postProcessAlgorithm(self, context, feedback):   
    return {}

  def displayName(self):
    return self.tr("Cross section")

  def group(self):
    return self.tr('Set receivers')

  def groupId(self):
    return 'receiver'

  def createInstance(self):
    return receivercrosssection()
